# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Literal

SessionStatus = Literal["paused", "running", "completed", "compacting", "compacted", "restoring", "restoration_failed"]
"""
Session status values.
"""
